/*
 * blogd.c
 *
 * Copyright 2000 Werner Fink, 2000 SuSE GmbH Nuernberg, Germany.
 *
 * This source is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifndef  _GNU_SOURCE
# define _GNU_SOURCE
#endif
#ifndef  BOOT_LOGFILE
# define BOOT_LOGFILE		"/var/log/boot.msg"
#endif
#include <sys/time.h>
#include <sys/types.h> /* Defines the macros major and minor */
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <time.h>
#include <pty.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stropts.h>
#include <dirent.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <limits.h>
#include <errno.h>
#include <paths.h>
#include <linux/major.h>
#include "libconsole.h"
#ifndef  _POSIX_MAX_CANON
# define _POSIX_MAX_CANON 255
#endif
extern volatile sig_atomic_t nsigsys;
extern volatile sig_atomic_t signaled;

static const char console[] = "/dev/console";

/*
 * Internal logger
 */
static char *myname = NULL;
static void _logger (const char *fmt, va_list ap)
{
    char buf[strlen(myname)+2+strlen(fmt)+1];
    strcat(strcat(strcpy(buf, myname), ": "), fmt);
    vfprintf(stderr, buf, ap);
    return;
}

/*
 * Cry and exit.
 */
void error (const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    _logger(fmt, ap);
    va_end(ap);
    popd();
    exit(EXIT_FAILURE);
}

/*
 * Warn the user.
 */
void warn (const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    _logger(fmt, ap);
    va_end(ap);
    return;
}

/*
 * Remove pidfile
 */
static void rmfpid()
{
    char buf[strlen(_PATH_VARRUN)+strlen(myname)+1+3+1];
    strcat(strcat(strcpy(buf, _PATH_VARRUN), myname), ".pid");
    unlink(buf);
}

/*
 * Write pidfile
 */
static void attribute((noinline)) pidfile()
{
    char buf[strlen(_PATH_VARRUN)+strlen(myname)+1+3+1];
    FILE * fpid;

    strcat(strcat(strcpy(buf, _PATH_VARRUN), myname), ".pid");
    if ((fpid = fopen (buf, "w")) == NULL) {
	warn("can not open %s: %m\n", buf);
	goto out;
    }
    fprintf(fpid, "%d\n", (int)getpid());
    fclose(fpid);
    atexit(rmfpid);
out:
    return;
}

/*
 *  Signal handler
 */
static struct sigaction saved_sigttin;
static struct sigaction saved_sigttou;
static struct sigaction saved_sigtstp;
static struct sigaction saved_sighup;
static struct sigaction saved_sigint;
static struct sigaction saved_sigquit;
static struct sigaction saved_sigterm;
static struct sigaction saved_sigsys;

static void sighandle(int sig)
{
    if (nsigsys && (sig == SIGTERM))
	return;
    signaled = (volatile sig_atomic_t)sig;
}

/*
 * Stop writing logs to disk, only repeat messages
 */
static void sigsys(int sig)
{
    nsigsys = (volatile sig_atomic_t)sig;
}

static void set_signal(int sig, struct sigaction *old, sighandler_t handler)
{
    do {
	if (sigaction(sig, NULL, old) == 0)
	    break;
    } while (errno == EINTR);

    if (old && old->sa_handler != handler) {
	struct sigaction new;
	sigset_t sigset;

	new.sa_handler = handler;
	sigemptyset(&new.sa_mask);
	new.sa_flags = SA_RESTART;
	do {
	    if (sigaction(sig, &new, NULL) == 0)
		break;
	} while (errno == EINTR);

	sigemptyset(&sigset);
	sigaddset(&sigset, sig);
	sigprocmask(SIG_UNBLOCK, &sigset, NULL);
    }
}

static void reset_signal(int sig, struct sigaction *old)
{
    struct sigaction cur;

    do {
	if (sigaction(sig, NULL, &cur) == 0)
	    break;
    } while (errno == EINTR);

    if (old && old->sa_handler == cur.sa_handler) {
	do {
	    if (sigaction(sig, old, NULL) == 0)
		break;
	} while (errno == EINTR);
    }
}

/*
 * To be able to reconnect to real tty on EIO
 */
static void reconnect(int fd)
{
    struct console * c;

    for (c = cons; c; c = c->next) {
	int newfd;

	if (!c->tty) continue;

	if (c->fd != fd) continue;

	switch (c->fd) {
	case 0:
	case -1:	/* Weired */
	    break;
	default:	/* IO of system consoles */
	    if ((newfd = open(c->tty, O_WRONLY|O_NONBLOCK|O_NOCTTY)) < 0)
		error("can not open %s: %m\n", c->tty);
	    dup2(newfd, c->fd);
	    if (newfd != c->fd)
		close(newfd);
	    break;
	}
    }
}


static volatile int cntrtty = 1;
static volatile pid_t pid = -1;
static volatile pid_t ppid;

static void flush_handler (void) attribute((noinline));
static void exit_handler (void) attribute((noinline));

/*
 * Now do the job
 */
int main(int argc, char *argv[])
{
    char ptsname[NAME_MAX+1];
    struct console *c;
    struct termios o;
    struct winsize w;
    struct stat st;
    speed_t ospeed;
    speed_t ispeed;
    int ptm, pts;
    int flags;
    pid_t pgrp;
    time_t tt;
    char *tty, *stt;

    ppid = getppid();
    pgrp = getpgid(ppid);
    myname = basename(*argv);

    if (stat(console, &st) < 0)
	error("No system console %s: %m\n", console);
    else {
	dev_t sysconsole = st.st_rdev;
	int fd0;

	if (stat("/proc/self/fd/0", &st) < 0)
	    error("No standard input: %m\n");
	if (sysconsole != st.st_rdev) {
	    close(0);
	    if ((fd0 = open(console, O_WRONLY|O_NONBLOCK|O_NOCTTY)) < 0)
		error("Can not open system console %s: %m\n", console);
	    if (fd0 > 0) {
		dup2(fd0, 0);
		close(fd0);
	    }
	}
    }

    if (argc > 2)
	error("Too many args; usage: %s [/dev/tty<X>]\n", myname);

    if (argc == 2)
	tty = argv[1];
    else
	tty = fetchtty(getpid(), ppid, NULL);

    if (!tty || !*tty)
	error("can not discover real system console tty, boot logging disabled.\n");

    if (strcmp(tty, console) == 0)
	error("console=%s, stdin=%s, must differ, boot logging disabled\n", tty, console);

    if (posix_memalign((void*)&cons, sizeof(void*), alignof(typeof(struct console))) != 0)
	error("memory allocation: %m\n");

    cons->next = (struct console*)0;
    cons->tty = tty;
    if ((cons->fd = open(cons->tty, O_WRONLY|O_NONBLOCK|O_NOCTTY)) < 0)
	error("can not open %s: %m\n", tty);

    if ((flags = fcntl(cons->fd, F_GETFL)) < 0)
	error("can not get terminal flags of %s: %m\n", tty);

    flags &= ~(O_NONBLOCK);
    flags |=   O_NOCTTY;
    if (fcntl(cons->fd, F_SETFL, flags) < 0)
	error("can not set terminal flags of %s: %m\n", tty);

    if (tcgetattr(cons->fd, &cons->otio) < 0)
	error("can not get terminal parameters of %s: %m\n", tty);

    ispeed = B38400;
    ospeed = B38400;
    if (ioctl(cons->fd, TIOCMGET, &flags) == 0) {
	ispeed = cfgetispeed(&cons->otio);
	ospeed = cfgetospeed(&cons->otio);
    }
    cfsetispeed(&cons->otio, ispeed);
    cfsetospeed(&cons->otio, ospeed);

    if (tcgetpgrp(cons->fd) < 0)
	cntrtty = 0;

    w.ws_row = 0;
    w.ws_col = 0;
    if (ioctl(cons->fd, TIOCGWINSZ, &w) < 0)
	error("can not get window size of %s: %m\n", tty);

    if (!w.ws_row)
	w.ws_row = 24;
    if (!w.ws_col)
	w.ws_col = 80;

    memcpy(&o, &cons->otio, sizeof(o));
    cfmakeraw(&o);
    cfsetispeed(&o, ispeed);
    cfsetospeed(&o, ospeed);
    o.c_lflag &= ~ECHO;
    o.c_lflag |= ISIG;
    o.c_cc[VTIME] = 0;
    o.c_cc[VMIN]  = CMIN;

    if (openpty(&ptm, &pts, ptsname, &o, &w) < 0)
	error("can not open pty/tty pair: %m\n");

    if (fstat(pts, &st) < 0)
	error("can not stat slave pty : %m\n");
    else {
	struct termios lock;
	memset(&lock, 0xff, sizeof(lock));
	(void)ioctl(pts, TIOCSLCKTRMIOS, &lock);
    }

    if (isastream(1) < 0 && errno == EBADF)
	dup2(0, 1);
    if (isastream(2) < 0 && errno == EBADF)
	dup2(0, 2);

    secondtty(cons, st.st_rdev);

    (void)ioctl(0, TIOCCONS, NULL);  /* Undo any current map if any */
    close(0);

    if (ioctl(pts, TIOCCONS, NULL) < 0)
	error("can not set console device to %s: %m\n", ptsname);
    close(pts);

    signaled = nsigsys = 0;
    set_signal(SIGTTIN, &saved_sigttin, SIG_IGN);
    set_signal(SIGTTOU, &saved_sigttou, SIG_IGN);
    set_signal(SIGTSTP, &saved_sigtstp, SIG_IGN);
    set_signal(SIGHUP,  &saved_sighup,  SIG_IGN);
    set_signal(SIGINT,  &saved_sigint,  sighandle);
    set_signal(SIGQUIT, &saved_sigquit, sighandle);
    set_signal(SIGTERM, &saved_sigterm, sighandle);
    set_signal(SIGSYS,  &saved_sigsys,  sigsys);
    (void)siginterrupt(SIGINT,  0);
    (void)siginterrupt(SIGQUIT, 0);
    (void)siginterrupt(SIGTERM, 0);
    (void)siginterrupt(SIGSYS,  0);

    for (c = cons; c; c = c->next) {
	int iflag, oflag;
#ifdef _PC_MAX_CANON
	if ((c->max_canon = (ssize_t)fpathconf(c->fd, _PC_MAX_CANON)) <= 0)
#endif
	    c->max_canon = _POSIX_MAX_CANON;
	c->tlock = 0;
	if (tcgetattr(c->fd, &c->otio) < 0)
	    continue;
	c->tlock = 1;

	iflag = c->otio.c_iflag;
	oflag = c->otio.c_oflag;

	if (ioctl(c->fd, TIOCMGET, &flags) == 0) {
		ispeed = cfgetispeed(&c->otio);
		ospeed = cfgetospeed(&c->otio);

		c->otio.c_iflag = c->otio.c_lflag = 0;
		c->otio.c_oflag = (ONLCR | OPOST);
		c->otio.c_cflag = CREAD | CS8 | HUPCL | (c->otio.c_cflag & CLOCAL);

		cfsetispeed(&c->otio, ispeed);
		cfsetospeed(&c->otio, ospeed);
	} else {
		c->otio.c_iflag |= (ICRNL | IGNBRK);
		c->otio.c_iflag &= ~(INLCR | IGNCR | BRKINT);
		c->otio.c_oflag |= (ONLCR | OPOST);
		c->otio.c_oflag &= ~(OCRNL | ONLRET);
	}
	(void)tcsetattr(c->fd, TCSADRAIN, &c->otio);

	c->otio.c_iflag = iflag;
	c->otio.c_oflag = oflag;

	if ((c->otio.c_lflag & ICANON) == 0) {
	    c->otio.c_lflag |= ICANON | IEXTEN | ISIG | ECHO|ECHOE|ECHOK|ECHOKE;
	    c->otio.c_oflag |= OPOST;
	    c->otio.c_iflag |= BRKINT;
	    c->otio.c_iflag &= ~IGNBRK;
	}
	if (ioctl(c->fd, TIOCGLCKTRMIOS, &c->ltio) == 0) {
	    /*
	     * Remove any lock as e.g. startpar(8) will use it
	     */
	    struct termios lock;
	    memset(&lock, 0x00, sizeof(lock));
	    if (ioctl(c->fd, TIOCSLCKTRMIOS, &lock) == 0)
		c->tlock = 2;
	}
    }

    atexit(exit_handler);	/* Register main exit handler */

    switch ((pid = fork())) {
    case 0:
	/* Get our own session */
	setsid();
	/* Reconnect our own terminal I/O */
	dup2(ptm, 0);
	dup2(cons->fd,  1);
	dup2(cons->fd,  2);
	if (cons->fd > 2)
	    dup2(1, cons->fd);	/* Choose stdout */
	close(ptm);
	break;
    case -1:
	close(ptm);
	error("can not fork to become daemon: %m\n");
    default:
	time(&tt);
	stt = ctime(&tt);
	close(ptm);
	for (c = cons; c; c = c->next) {
	    if (c->fd > 0) {
		close(c->fd);
		c->fd = -1;
	    }
	}
	fprintf(stdout, "\rBoot logging started on %s(%s) at %.24s\n", tty, console, stt);
	fflush(stdout);
	_exit(EXIT_SUCCESS);	/* NEVER rise exit handlers here */
    }

    atexit(flush_handler);	/* Register flush exit handler */

    prepareIO(reconnect, pidfile, pgrp, 0);
    while (!signaled)
	safeIO();

    exit(EXIT_SUCCESS);		/* Raise exit handlers */
}

static void flush_handler (void)
{
    (void)tcdrain(1);
    (void)tcdrain(2);

    if (!cntrtty) kill(ppid, SIGSTOP);
    closeIO();
    if (!cntrtty) kill(ppid, SIGCONT);
}

static void exit_handler (void)
{
    struct console *c;
    int fd;

    for (c = cons; c; c = c->next) {
	if (c->fd > 0) {
	    if (c->tlock)	/* write back old setup  */
		tcsetattr(c->fd, TCSADRAIN, &c->otio);
	    if (c->tlock > 1)	/* write back lock if any */
		(void)ioctl(c->fd, TIOCSLCKTRMIOS, &c->ltio);
	    c->tlock = 0;
	    close(c->fd);
	    c->fd = -1;
	}
    }

    errno = 0;
    if ((fd = open(console, O_RDWR|O_NOCTTY)) >= 0) {
	(void)ioctl(fd, TIOCCONS, NULL);	/* Restore old console mapping */
	close(fd);
    }

    close(1);
    close(2);
    (void)tcflush(0, TCIFLUSH);
    close(0);

    reset_signal(SIGTTIN, &saved_sigttin);
    reset_signal(SIGTTOU, &saved_sigttou);
    reset_signal(SIGTSTP, &saved_sigtstp);
    reset_signal(SIGHUP,  &saved_sighup);
    reset_signal(SIGINT,  &saved_sigint);
    reset_signal(SIGQUIT, &saved_sigquit);
    reset_signal(SIGTERM, &saved_sigterm);
    reset_signal(SIGSYS,  &saved_sigsys);
}
