/*
 * libconsole.c
 *
 * Copyright 2000 Werner Fink, 2000 SuSE GmbH Nuernberg, Germany.
 *
 * This source is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#undef BLOGD_EXT
#ifndef  _GNU_SOURCE
# define _GNU_SOURCE
#endif
#ifndef  LOG_BUFFER_SIZE
# define LOG_BUFFER_SIZE	65536
#endif
#ifndef  TRANS_BUFFER_SIZE
# define TRANS_BUFFER_SIZE	 4096
#endif
#ifndef  BOOT_LOGFILE
# define BOOT_LOGFILE		"/var/log/boot.msg"
#endif
#ifndef  _PATH_BLOG_FIFO
# define _PATH_BLOG_FIFO	"/dev/blog"
#endif
#include <sys/time.h>
#include <sys/types.h> /* Defines the macros major and minor */
#include <sys/stat.h>
#include <sys/ioctl.h>
#ifndef TIOCGDEV
# include <sys/vt.h>
# include <sys/kd.h>
# include <linux/serial.h>
#endif
#include <linux/major.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/un.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <dirent.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <signal.h>
#include <limits.h>
#include <errno.h>
#include <pthread.h>
#include <sched.h>
#include <syscall.h>
#include "libconsole.h"
#include "listing.h"

/*
 * Used to ignore some signals during pselect(2)
 */
static sigset_t omask;

/*
 * Remember if we're signaled.
 */
volatile sig_atomic_t signaled;

/*
 * Error raised in exit handler should not call exit(3) its self
 */
#define lerror(fmt, args...)	    \
    do {			    \
	if (signaled) {		    \
	    warn(fmt, args);	    \
		goto out;	    \
	} error(fmt, args);	    \
    } while (0)

/*
 * push and popd direcotry changes
 */

typedef struct pwd_struct {
    list_t      deep;
    char        *pwd;
} pwd_t;
#define getpwd(list)	list_entry((list), struct pwd_struct, deep)
static list_t pwd = { &(pwd), &(pwd) }, * topd = &(pwd);

void pushd(const char * path)
{
    pwd_t *  dir;

    dir = (pwd_t *)malloc(sizeof(pwd_t));
    if (dir) {
	if (!(dir->pwd = getcwd(NULL, 0)))
	    goto err;
	insert(&(dir->deep), topd->prev);
	goto out;
    }
err:
    error("%m");
out:
    if (chdir(path) < 0)
	error ("pushd() can not change to directory %s: %m\n", path);
}

void popd(void)
{
    list_t * tail = topd->prev;
    pwd_t *  dir;

    if (list_empty(topd))
	goto out;
    dir = getpwd(tail);
    if (chdir(dir->pwd) < 0)
	error ("popd() can not change directory %s: %m\n", dir->pwd);
    free(dir->pwd);
    delete(tail);
    free(dir);
out:
    return;
}

/*
 * Follow the link to its full deep, this because
 * to get the real file name stored in lnk[].
 */
static char lnk[PATH_MAX+1];
static int rlstat(char ** file, struct stat *st)
{
    int ret = -1;
    unsigned int deep = 0;
    static ino_t iprev;			/* Recurive loop detection */

    if (lstat(*file, st) < 0)
	goto out;

    if (iprev && iprev == st->st_ino)	/* We're run on the previous */
	goto out;
    iprev = st->st_ino;

    do {
	const char *prev_file;
	ssize_t cnt;

	if (!S_ISLNK(st->st_mode))	/* Not a symlink */
	    break;

	if (deep++ > MAXSYMLINKS) {
	    errno = EMLINK;
	    goto out;
	}

	prev_file = strdupa(*file);
	if (!prev_file)
	    error("strdupa(): %m\n");
	if ((cnt = readlink(*file, lnk, sizeof(lnk) - 1)) < 0)
	    goto out;
	lnk[cnt] = '\0';
	if (lnk[0] != '/') {
	    /* Construct a new valid file name */
	    const char *lastslash;

	    if ((lastslash = strrchr (prev_file, '/'))) {
		size_t dirname_len = lastslash - prev_file + 1;
		if (dirname_len + cnt > PATH_MAX)
		    cnt = PATH_MAX - dirname_len;
		memmove(&lnk[dirname_len], &lnk[0], cnt + 1);
		memcpy(&lnk[0], prev_file, dirname_len);
	    }
	}
	*file = lnk;

	if (lstat(*file, st) < 0)
	    goto out;

    } while (S_ISLNK(st->st_mode));
    ret = 0;
    if (*file != (void *)&lnk) {
	strcpy(lnk, *file);
	*file = lnk;
    }
out:
    return ret;
}

/*
 * Arg used: safe out
 */
static void (*vc_reconnect)(int fd);
static inline void safeout (int fd, const char *ptr, size_t s, ssize_t max)
{
    int saveerr = errno;
    int repeated = 0;
    static int eiocount;

    while (s > 0) {
	ssize_t p = write (fd, ptr, (max < 1) ? 1 : ((s < (size_t)max) ? s : (size_t)max));
	if (p < 0) {
	    if (errno == EPIPE) {
		warn("error on writing to fd %d: %m\n", fd);
		if (signaled)
		    goto out;
		exit(EXIT_SUCCESS);
	    }
	    if (errno == EINTR) {
		errno = 0;
		continue;
	    }
	    if (errno == EAGAIN) {	/* Kernel 2.6 seems todo this very often */
		int ret;
		fd_set check;

		if (repeated++ > 1000) {
		    lerror("repeated error on writing to fd %d: %m\n", fd);
		}

		FD_ZERO (&check);
		FD_SET (fd, &check);

		/* Avoid high load: wait upto two seconds if system is not ready */
		errno = 0;
		do {
		    struct timespec two = {2, 0};
		    ret = pselect(fd + 1, (fd_set*)0, &check, (fd_set*)0, &two, &omask);

		} while ((ret < 0) && (errno == EINTR));

		if (ret < 0) {
		    lerror("can not write to fd %d: %m\n", fd);
		}

		errno = 0;
		continue;
	    }
	    if (errno == EIO) {
		if ((eiocount++ > 10) || !vc_reconnect) {
		    lerror("can not write to fd %d: %m\n", fd);
		}
		(*vc_reconnect)(fd);
		errno = 0;
		continue;
	    }
	    lerror("can not write to fd %d: %m\n", fd);
	}
	repeated = 0;
	ptr += p;
	s -= p;
    }
out:
    errno = saveerr;
}

/*
 * Twice used: safe in
 */
static inline ssize_t safein (int fd, char *ptr, size_t s, const int noerr)
{
    int saveerr = errno;
    ssize_t r = 0;
    int t;
    static int repeated;

    if (s > SSIZE_MAX)
	s = SSIZE_MAX;

    t=0;
    if ((ioctl(fd, FIONREAD, &t) < 0) || (t == 0)) {
	fd_set check;
	struct timespec zero = {0, 0};

	do {
	    FD_ZERO (&check);
	    FD_SET (fd, &check);

	    /* Avoid deadlock: do not read if nothing is in there */
	    if (pselect(fd + 1, &check, (fd_set*)0, (fd_set*)0, &zero, &omask) <= 0)
		break;

	    r = read (fd, ptr, s);

	} while (r < 0 && (errno == EINTR || errno == EAGAIN));

	/* Do not exit on a broken FIFO */
	if (r < 0 && errno != EPIPE) {
	    if (noerr || signaled)
		goto out;
	    if (fd == 0 && errno == EIO)
		warn("\e[31m\e[1msystem console stolen at line %d!\e[m\n", __LINE__);
	    lerror("Can not read from fd %d: %m\n", fd);
	}

	goto out;
    }

    if (t > 0 && (size_t)t > s)
	t = s;

    repeated = 0;
    while (t > 0) {
	ssize_t p = read (fd, ptr, t);
	if (p < 0) {
	    if (repeated++ > 1000) {
		lerror("Repeated error on reading from fd %d: %m\n", fd);
	    }
	    if (errno == EINTR || errno == EAGAIN) {
		errno = 0;
		continue;
	    }
	    if (noerr || signaled)
		goto out;
	    if (fd == 0 && errno == EIO)
		warn("\e[31m\e[1msystem console stolen at line %d!\e[m\n", __LINE__);
	    lerror("Can not read from fd %d: %m\n", fd);
	}
	repeated = 0;
	ptr += p;
	r += p;
	t -= p;
    }
out:
    errno = saveerr;
    return r;
}

/*
 * The stdio file pointer for our log file
 */
struct console *cons;
static FILE * flog = NULL;
static int fdread  = -1;
static int fdfifo  = -1;

/*
 * Signal control for writing on log file
 */
volatile sig_atomic_t nsigsys;
static volatile sig_atomic_t nsigio = -1;
static sigset_t save_oldset;

/* One shot signal handler */
static void sigio(int sig)
{
    /* Currently no thread active */
    if (nsigio == 0) {
	(void)sigprocmask(SIG_SETMASK, &save_oldset, NULL);
	(void)signal(sig, SIG_IGN);
    }
    nsigio = sig;
}

/*
 * Our ring buffer
 */

typedef struct _mutex {
    int locked;
    pthread_mutex_t mutex;
    pthread_t thread;
} mutex_t;

static inline void lock(mutex_t *mutex)
{
    if (mutex->thread != pthread_self() || !mutex->locked) {
	pthread_mutex_lock(&mutex->mutex);
	mutex->thread = pthread_self();
    }
    mutex->locked++;
}

static inline void unlock(mutex_t *mutex)
{
    if (!--mutex->locked) {
	mutex->thread = 0;
	pthread_mutex_unlock(&mutex->mutex);
    }
}

static mutex_t llock = { 0, PTHREAD_MUTEX_INITIALIZER, 0 };
static mutex_t ljoin = { 0, PTHREAD_MUTEX_INITIALIZER, 0 };
static pthread_cond_t lcond = PTHREAD_COND_INITIALIZER;
static pthread_t    lthread;
static volatile int running;

static       unsigned char data[LOG_BUFFER_SIZE];
static const unsigned char *const end = data + sizeof(data);
static       unsigned char *     head = data;
static       unsigned char *     tail = data;
static volatile ssize_t avail;
#define THRESHOLD	64

static inline void resetlog(void) { tail = head = data; avail = 0; }

static inline void storelog(const unsigned char *const buf, const size_t len)
{
    if (len > (size_t)(end - tail)) {
	static int be_warned = 0;
	if (!be_warned) {
	    warn("log buffer exceeded\n");
	    be_warned++;
	}
	goto xout;
    }
    memcpy(tail, buf, len);
    avail = (tail += len) - head;
xout:
    return;
}

static inline void addlog(const unsigned char c)
{
    if (end - tail <= 0) {
	static int be_warned = 0;
	if (!be_warned) {
	    warn("log buffer exceeded\n");
	    be_warned++;
	}
	goto xout;
    }
    *tail = c;
    avail = (tail += 1) - head;
xout:
    return;
}

static inline void writelog(void)
{
    if (!flog)
	return;
    clearerr(flog);
    lock(&llock);
    while (avail > 0) {
	size_t ret = (size_t)avail;

	if (avail > TRANS_BUFFER_SIZE)
	    ret = TRANS_BUFFER_SIZE;

	if (!flog || nsigsys)
	    break;
	ret = fwrite(head, sizeof(unsigned char), ret, flog);
	if (!ret && ferror(flog))
	    break;
	head += ret;

	if (head >= tail) {		/* empty, reset buffer */
	    resetlog();
	    break;
	}
	if (head > data) {		/* buffer not empty, move contents */
	    avail = tail - head;
	    head = (unsigned char *)memmove(data, head, avail);
	    tail = head + avail;
	}
    }
    unlock(&llock);
    if (flog) {
	fflush(flog);
	fdatasync(fileno(flog));
    }
}

static inline void flushlog(void)
{
    if (running) pthread_cond_broadcast(&lcond);
}

static inline int poll(int msec, mutex_t *outer)
{
    int ret = 1;

    if (avail <= THRESHOLD) {
	struct timeval now;

	if (gettimeofday(&now, NULL) == 0) {
	    struct timespec abstime;
	    int err;

	    now.tv_usec += msec * 1000;
	    while (now.tv_usec >= 1000000) {
		now.tv_sec++;
		now.tv_usec -= 1000000;
	    }
	    abstime.tv_sec  = now.tv_sec;
	    abstime.tv_nsec = now.tv_usec * 1000;

	    do {
		int locked = outer->locked;
		/* Note: pthread_cond_timedwait() atomically unlocks the mutex */
		outer->locked = 0;
		err = pthread_cond_timedwait(&lcond, &outer->mutex, &abstime);
		outer->locked = locked;
	    } while (err == EINTR);

	    if (err == ETIMEDOUT || err == EBUSY)
		ret = 0;
	}
    } else
	pthread_yield();

    return ret;
}

/*
 * Our transfer buffer
 */
static unsigned char trans[TRANS_BUFFER_SIZE];

/*
 * Remove Escaped sequences and write out result (see
 * linux/drivers/char/console.c in do_con_trol()).
 */
enum {	ESnormal, ESesc, ESsquare, ESgetpars, ESgotpars, ESfunckey,
	EShash, ESsetG0, ESsetG1, ESpercent, ESignore, ESnonstd,
	ESpalette };
#define NPAR 16
static unsigned int state = ESnormal;
static int npar, nl;
static unsigned long int line;

/*
 * Workaround for spinner of fsck/e2fsck
 * Uses ascii lines ending with '\r' only
 */
static int spin;
#define PROGLEN 192
static unsigned char prog[PROGLEN];

static void parselog(unsigned char *buf, const size_t s)
{
    int c;
    ssize_t r = s, up;
    unsigned char uprt[16];

    lock(&llock);

    while (r > 0) {
	c = (unsigned char)*buf;

	switch(state) {
	case ESnormal:
	default:
	    state = ESnormal;
	    switch (c) {
	    case 0  ...  8:
	    case 16 ... 23:
	    case 25:
	    case 28 ... 31:
		nl = 0;
		spin = 0;
		addlog('^'); addlog(c + 64);
		break;
	    case '\n':
		if (spin > 4)	/* last spinner line */
		    storelog(prog, strlen((char*)prog));
		nl = 1;
		line++;
		spin = 0;
		addlog(c);
		break;
	    case '\r':
		spin++;
		if (spin < 5) {
		    if (spin > 1)
			addlog('\n');
		    nl = 1;
		}
		if (spin == 5)
		    storelog((unsigned char*)"\n<progress bar skipped>\n", 24);
		break;
	    case 14:
	    case 15:
		/* ^N and ^O used in xterm for rmacs/smacs  *
		 * on console \033[10m and \033[11m is used */
	    case 24:
	    case 26:
		spin = 0;
		break;
	    case '\033':
		spin = 0;
		state = ESesc;
		break;
	    case '\t':
	    case  32 ... 126:
	    case 160 ... 255:
		if (spin < 5) {
		    if (spin == 1 && nl)
			addlog('\n');
		    addlog(c);
		} else {		/* Seems to be a lengthy spinner line */
		    static   int old = 0;
		    static ssize_t p = 0;
		    if (old != spin) {
			old = spin;	/* Next line overwrite on tty */
			p = 0;
			bzero(prog, PROGLEN);
		    }
		    if (p < PROGLEN)
			prog[p++] = c;	/* buffer always current line */
		}
		nl = 0;
		break;
	    case 127:
		nl = 0;
		spin = 0;
		addlog('^'); addlog('?');
		break;
	    case 128 ... 128+26:
	    case 128+28 ... 159:
		nl = 0;
		spin = 0;
		if ((up = snprintf((char*)uprt, sizeof(uprt), "\\%03o", c)) > 0)
		    storelog(uprt, (size_t)up);
		break;
	    case 128+27:
		spin = 0;
		state = ESsquare;
		break;
	    default:
		nl = 0;
		spin = 0;
		if ((up = snprintf((char*)uprt, sizeof(uprt), "0x%X", c)) > 0)
		    storelog(uprt, (size_t)up);
		break;
	    }
	    break;
	case ESesc:
	    state = ESnormal;
	    switch((unsigned char)c) {
	    case '[':
		state = ESsquare;
		break;
	    case ']':
		state = ESnonstd;
		break;
	    case '%':
		state = ESpercent;
		break;
	    case 'E':
	    case 'D':
		if (spin > 4)	/* last spinner line */
		    storelog(prog, strlen((char*)prog));
		nl = 1;
		line++;
		spin = 0;
		addlog('\n');
		break;
	    case '(':
		state = ESsetG0;
		break;
	    case ')':
		state = ESsetG1;
		break;
	    case '#':
		state = EShash;
		break;
#ifdef BLOGD_EXT
	    case '^':				/* Boot log extension */
		state = ESignore;
		break;
#endif
	    default:
		break;
	    }
	    break;
	case ESnonstd:
	    if        (c == 'P') {
		npar = 0;
		state = ESpalette;
	    } else if (c == 'R')
		state = ESnormal;
	    else
		state = ESnormal;
	    break;
	case ESpalette:
	    if ((c>='0'&&c<='9') || (c>='A'&&c<='F') || (c>='a'&&c<='f')) {
		npar++;
		if (npar==7)
		    state = ESnormal;
	    } else
		state = ESnormal;
	    break;
	case ESsquare:
	    npar = 0;
	    state = ESgetpars;
	    if (c == '[') {
		state = ESfunckey;
		break;
	    }
	    if (c == '?')
		break;
	case ESgetpars:
	    if (c==';' && npar<NPAR-1) {
		npar++;
		break;
	    } else if (c>='0' && c<='9') {
		break;
	    } else
		state = ESgotpars;
	case ESgotpars:
	    state = ESnormal;
	    break;
	case ESpercent:
	    state = ESnormal;
	    break;
	case ESfunckey:
	case EShash:
	case ESsetG0:
	case ESsetG1:
	    state = ESnormal;
	    break;
#ifdef BLOGD_EXT
	case ESignore:				/* Boot log extension */
	    state = ESesc;
	    {
		unsigned char echo[64];
		ssize_t len;

		/* Release the lock while doing IO */
		unlock(&llock);
		if ((len = snprintf(echo, sizeof(echo), "\033[%lu;%dR", line, nl)) > 0)
		    safeout(fdread, echo, len, -1);
		else
		    safeout(fdread, "\033R", 2, -1);
		tcdrain(fdread);
		lock(&llock);
	    }
	    break;
#endif
	}
	buf++;
	r--;
    }

    unlock(&llock);
}

static void copylog(const unsigned char *buf, const size_t s)
{
    lock(&llock);
    if (!nl)
	addlog('\n');
    storelog(buf, s);
    if (buf[s-1] != '\n')
	addlog('\n');
    nl = 1;
    unlock(&llock);
}

static void *action(void *dummy attribute((unused)))
{
    sigset_t sigset, save_oldset;
    sigemptyset(&sigset);

    sigaddset(&sigset, SIGTTIN);
    sigaddset(&sigset, SIGTTOU);
    sigaddset(&sigset, SIGTSTP);
    sigaddset(&sigset, SIGHUP);
    sigaddset(&sigset, SIGINT);
    sigaddset(&sigset, SIGQUIT);
    sigaddset(&sigset, SIGTERM);
    sigaddset(&sigset, SIGSYS);
    (void)pthread_sigmask(SIG_BLOCK, &sigset, &save_oldset);

    lock(&ljoin);
    running = 1;
    while (running) {

	if (!poll(30, &ljoin))
	    continue;

	if (!flog)
	    break;

	writelog();
    }
    unlock(&ljoin);

    (void)pthread_sigmask(SIG_SETMASK, &save_oldset, NULL);
    return NULL;
}

/*
 * Prepare I/O
 */
static const char *fifo_name = _PATH_BLOG_FIFO;
static void (*vr_access)(void);
static pid_t pgroup = -1;

void prepareIO(void (*rfunc)(int), void (*pfunc)(void), const pid_t pgrp, const int in)
{
    vc_reconnect = rfunc;
    vr_access = pfunc;
    pgroup  = pgrp;
    fdread  = in;

    (void)sigfillset(&omask);
    (void)sigdelset(&omask, SIGQUIT);
    (void)sigdelset(&omask, SIGTERM);
    (void)sigdelset(&omask, SIGSYS);
    (void)sigdelset(&omask, SIGIO);

    if (fifo_name && fdfifo < 0) {
	struct stat st;
	errno = 0;
	/* udev support: create /dev/blog if not exist */
	if (stat(fifo_name, &st) < 0 && errno == ENOENT)
	    (void)mkfifo(fifo_name, 0600);
	errno = 0;
	if (!stat(fifo_name, &st) && S_ISFIFO(st.st_mode)) {
	    if ((fdfifo = open(fifo_name, O_RDWR|O_NOCTTY|O_CLOEXEC)) < 0)
		warn("can not open named fifo %s: %m\n", fifo_name);
	}
    }
    (void)mlockall(MCL_FUTURE);
}

/*
 * Seek for input, more input ...
 */
static void more_input (struct timespec *timeout, const int noerr)
{
    fd_set watch;
    int nfds, wfds;

    FD_ZERO (&watch);
    FD_SET (fdread, &watch);

    if (fdfifo >= 0) {
	FD_SET (fdfifo, &watch);
	wfds = (fdread > fdfifo ? fdread : fdfifo) + 1;
    } else
	wfds = fdread + 1;

    nfds = pselect(wfds, &watch, (fd_set*)0, (fd_set*)0, timeout, &omask);

    if (nfds <= 0) {
	timeout->tv_sec  = 0;
	timeout->tv_nsec = 0;
	if (nfds < 0 && errno != EINTR && !signaled)
	    error ("pselect(): %m\n");
	goto nothing;
    }

    if (FD_ISSET(fdread, &watch)) {
	const ssize_t cnt = safein(fdread, (char*)trans, sizeof(trans), noerr);
	static struct winsize owz;
	struct winsize wz;

	if (cnt > 0) {
	    struct console * c;

	    if (ioctl(cons->fd, TIOCGWINSZ, &wz) == 0) {
		if (memcmp(&owz, &wz, sizeof(struct winsize))) {
		    ioctl(fdread, TIOCSWINSZ, &wz);
		    (void)memcpy(&owz, &wz, sizeof(struct winsize));
		    if (pgroup > 1)
		        killpg(pgroup, SIGWINCH);
		}
	    }
	    parselog(trans, cnt);			/* Parse and make copy of the input */

	    for (c = cons; c; c = c->next) {
		if (c->fd < 0)
		    continue;
		safeout(c->fd, (char*)trans, cnt, c->max_canon);
		(void)tcdrain(c->fd);			/* Write copy of input to real tty */
	    }

	    flushlog();
	}
    }

    if (fdfifo >= 0 && FD_ISSET(fdfifo, &watch)) {
	const ssize_t cnt = safein(fdfifo, (char*)trans, sizeof(trans), noerr);

	if (cnt > 0) {
	    copylog(trans, cnt);		/* Make copy of the input */
	    flushlog();
	}
    }
    errno = 0;
nothing:
    return;
}

/*
 *  The main routine for blogd.
 */
void safeIO (void)
{
    struct timespec timeout;
    static int log = -1;

    if (!nsigio) /* signal handler set but no signal recieved */
	goto skip;

    if (log < 0) {
#ifdef DEBUG_SIGIO
	if (nsigio < 0)
	   goto skip;
#else
	if (nsigio < 0) {
	    /*
	     * Maybe access() does lie under kernel 2.6
	     * or the root file system is rw mounted
	     */
	    const char *runlevel = getenv("RUNLEVEL");
	    if (runlevel && (*runlevel < '0' || *runlevel > '6'))
		goto skip;
	}
#endif
	if (access(BOOT_LOGFILE, W_OK) < 0) {
	    if (errno != ENOENT && errno != EROFS)
		error("Can not write to %s: %m\n", BOOT_LOGFILE);
	    goto skip;
	}
	if ((log = open(BOOT_LOGFILE, O_WRONLY|O_NOCTTY|O_APPEND)) < 0) {
	    if (errno != ENOENT && errno != EROFS)
		error("Can not open %s: %m\n", BOOT_LOGFILE);
	    goto skip;
	}
	if (vr_access) {
	    (*vr_access)();
	    vr_access = NULL;
	}
	lock(&llock);
	if ((flog = fdopen (log, "a")) == NULL) {
	    unlock(&llock);
	    error("Can not open %s: %m\n", BOOT_LOGFILE);
	}
	unlock(&llock);

	nsigio = SIGIO; /* We do not need a signal handler */
	(void)signal(SIGIO, SIG_IGN);
    }

skip:
    if (nsigio < 0) { /* signal handler not set, so do it */
	struct sigaction new;
	sigset_t sigset;

	/* Currently no thread active */
	nsigio = 0;

	new.sa_handler = sigio;
	sigemptyset(&new.sa_mask);
	new.sa_flags = SA_RESTART;
	do {
	    if (sigaction(SIGIO, &new, NULL) == 0)
		break;
	} while (errno == EINTR);

	sigemptyset(&sigset);
	sigaddset(&sigset, SIGIO);
	(void)sigprocmask(SIG_UNBLOCK, &sigset, &save_oldset);
    }

    timeout.tv_sec  = 5;
    timeout.tv_nsec = 0;
    more_input(&timeout, 0);

    if (flog && !running) {
	int policy = SCHED_RR;
	struct sched_param param;

	pthread_getschedparam(pthread_self(), &policy, &param);
	pthread_create(&lthread, NULL, &action, NULL);

	policy = SCHED_RR;
	param.sched_priority = sched_get_priority_max(policy)/2 + 1;
	pthread_setschedparam(pthread_self(), policy, &param);
	pthread_setschedparam(lthread, policy, &param);
    }

    if (nsigsys) {  /* Stop writing logs to disk, only repeat messages */
	if (flog)
	    closeIO();
	if (nsigio < 0) {
	    nsigio = SIGIO;
	    (void)signal(SIGIO, SIG_IGN);
	}
    }
}

/*
 *
 */
void closeIO(void)
{
    struct timespec timeout;
    struct console * c;
    int n = 240;

    /* Maybe we've catched a signal, therefore */
    if (!flog && !nsigsys)
	warn("no message logging because /var file system is not accessible\n");

    for (c = cons; c; c = c->next) {
	if (c->fd < 0)
	    continue;
	(void)tcdrain(c->fd);		/* Hold in sync with console */
    }

    do {
	/*
	 * Repeat this as long as required,
	 * but not more than 3 seconds
	 */

	if (!n)
	    break;
	n--;

	timeout.tv_sec  = 0;
	timeout.tv_nsec = 25*1000*1000;

    	more_input(&timeout, 1);
	(void)tcdrain(fdread);

    } while (timeout.tv_sec || timeout.tv_nsec);

    if (running) {
	lock(&ljoin);
	running = 0;
	unlock(&ljoin);
	flushlog();
	pthread_cancel(lthread);
    }

    if (flog) {
	writelog();
	if (!nl)
	    fputc('\n', flog);
	(void)fclose(flog);
    }
    lock(&llock);
    flog = NULL;
    unlock(&llock);

    if (fdfifo >= 0) {
	close(fdfifo);
	fdfifo = -1;
    }
    for (c = cons; c; c = c->next) {
	if (c->fd < 0)
	    continue;
	(void)tcdrain(c->fd);
    }
    return;
}

/*
 * Fetch our real underlying terminal character device,
 * returned name should be freed if not used anymore.
 */
static void ctty(pid_t pid, unsigned int * tty, pid_t * ttypgrp)
{
    char fetched[NAME_MAX+1];
    ssize_t cnt;
    int fd;

    sprintf(fetched, "/proc/%d/stat", (int)pid);
    if ((fd = open(fetched, O_RDONLY|O_NOCTTY)) < 0)
	error("can not open(%s): %m\n", fetched);
    cnt = safein(fd, fetched, sizeof(fetched), 0);
    close(fd);

    if (cnt <= 0)
	error("can not read from proc: %m\n");
    else	
	/* format pid comm state ppid pgrp session tty tpgid */
	if (sscanf(fetched,"%*d %*s %*c %*d %*d %*d %u %u %*u", tty, ttypgrp) != 2)
	    error("can not sscanf for my tty: %m\n");
}

/* fall back routine to fetch tty */
static dev_t fallback(const pid_t pid, const pid_t ppid)
{
    unsigned int tty = 0;
    pid_t ttypgrp = -1;
    pid_t  pgrp = getpgid(pid);
    pid_t ppgrp = getpgid(ppid);

    ctty(pid, &tty, &ttypgrp);

    if (pgrp != ttypgrp && ppgrp != ttypgrp) {
	int fdfrom[2];
	pid_t pid = -1;  /* Inner pid */

	if (pipe(fdfrom) < 0)
	    error("can not create a pipe: %m\n");

	switch ((pid = fork())) {
	case 0:
	    {   void (*save_sighup);

		dup2( fdfrom[1], 1);
		close(fdfrom[1]);
		close(fdfrom[0]);

		pid = getpid();	/* our pid is not zero */

		if (pid != getsid(pid)) {
		    if (pid == getpgid(pid))
			setpgid(0, ppgrp);
		    setsid();
		}

		/* Remove us from any controlling tty */
		save_sighup = signal(SIGHUP, SIG_IGN);
		if (ttypgrp > 0)
		    ioctl(0, TIOCNOTTY, (void *)1);
		(void)signal(SIGHUP, save_sighup);

		/* Take stdin as our controlling tt< */
		if (ioctl(0, TIOCSCTTY, (void *)1) < 0)
		    warn("can not set controlling tty: %m\n");

		ctty(pid, &tty, &ttypgrp);

		/* Never hold this controlling tty */
		save_sighup = signal(SIGHUP, SIG_IGN);
		if (ttypgrp > 0)
		    ioctl(0, TIOCNOTTY, (void *)1);
		(void)signal(SIGHUP, save_sighup);

		printf("|%u|%u|", tty, ttypgrp);  /* stdout to pipe synchronize ... */

		exit(EXIT_SUCCESS);
	    } break;
	case -1:
	    error("can not execute: %m\n");
	    break;
	default:
	    {   int fd = dup(0);
		dup2( fdfrom[0], 0);
		close(fdfrom[0]);
		close(fdfrom[1]);

		scanf("|%u|%u|", &tty, &ttypgrp); /* ... with stdin from pipe here  */

		dup2(fd, 0);
		close(fd);
	    } break;
	}
    }

    return (dev_t)tty;
}

static int checkdev(char ** retname, const dev_t dev, DIR * ddev)
{
    int found = 0, fd;
    struct dirent *d;
    static int deep;

    if (deep == 0) {
	
	fd = dirfd(ddev);
	rewinddir(ddev);

	while ((d = readdir(ddev))) {	/* First scan all devices direct in /dev/ of devtmpfs */
	    struct stat st;
	    char *name = d->d_name;
	    char path[PATH_MAX+1];
    
	    if (*name == '.')
		continue;
	    if (fstatat(fd, name, &st, 0) < 0)
		continue;
	    if (!S_ISCHR(st.st_mode))
		continue;
	    if (dev != st.st_rdev)
		continue;
	    if ((size_t)snprintf(path, sizeof(path), "/dev/%s", name) >= sizeof(path))
		continue;
	    name = realpath(path, NULL);
	    if (!name)
		break;
	    *retname = name;
	    found++;
	    break;
	}
	if (found)
	    goto out;
    }

    rewinddir(ddev);
    while ((d = readdir(ddev))) {	/* now allow also deeper entries of none devtmpfs */
	struct stat st;
	char *name = d->d_name;

	if (*name == '.')
	    continue;

	if (deep == 0) {
	    /*
	     * No terminals therein
	     */
	    if (!strcmp(name, "shm"))
		continue;

	    if (!strncmp(name, "snd", 3))
		continue;

	    if (!strncmp(name, "ram", 3))
		continue;

	    if (!strncmp(name, "rtc", 3))
		continue;

	    if (!strncmp(name, "card", 4))
		continue;

	    if (!strncmp(name, "vboxguest", 9))
		continue;

	    if (!strcmp(name, "cciss"))
		continue;

	    if (!strcmp(name, "ataraid"))
		continue;

	    if (!strcmp(name, "cbd"))
		continue;

	    if (!strcmp(name, "cpu"))
		continue;

	    if (!strcmp(name, "dvb"))
		continue;

	    if (!strcmp(name, "i2o"))
		continue;

	    if (!strcmp(name, "ida"))
		continue;

	    if (!strcmp(name, "inet"))
		continue;

	    if (!strcmp(name, "input"))
		continue;

	    if (!strcmp(name, "mvideo"))
		continue;

	    if (!strcmp(name, "raw"))
		continue;

	    if (!strcmp(name, "rd"))
		continue;

	    if (!strcmp(name, "scsi"))
		continue;

	    if (!strcmp(name, "block"))
		continue;

	    if (!strcmp(name, "cgroup"))
		continue;

	    if (!strcmp(name, "disk"))
		continue;

	    if (!strcmp(name, "bus"))
		continue;

	    if (!strcmp(name, "bsg"))
		continue;

	    if (!strcmp(name, "cpu"))
		continue;

	    if (!strcmp(name, "char"))
		continue;
	}

	if (rlstat(&name, &st) < 0) {
	    if (errno != ENOENT)
		warn("can not follow %s: %m\n", name);
	    continue;
	}

	if (!strncmp(name, "/proc", 5))
	    continue;

	if (S_ISDIR(st.st_mode)) {
	    DIR * cur = opendir(name);
	    if (!cur)
		error("can not opendir(%s): %m\n", name);
	    pushd(name);
	    deep++;
	    found = checkdev(retname, dev, cur);
	    deep--;
	    popd();
	    closedir(cur);

	    if (found)
		break;

	    continue;
	}

	if (!S_ISCHR(st.st_mode))
	    continue;

	if (dev != st.st_rdev)
	    continue;

	/*
	 * rlstat() uses lnk[] as buffer to which points
	 * the pointer name after rlstat().
	 */
	if (*name && *name != '/') {
	    char * ptr = strdupa(name);
	    if (!ptr)
		error("checkdev(): %m\n");
	    getcwd(name, PATH_MAX - 1 - strlen(ptr));
	    strcat(name, "/");
	    strcat(name, ptr);
	}

	found++;

	/*
	 * Allocate memory to be able to return several
	 * different buffers for different files names.
	 */
	name = realpath(name, NULL);
	if (!name)
	    error("checkdev(): %m\n");
	*retname = name;
	break;
    }
out:
    return found;
}

/* Read and allocate one line from file */
static char *oneline(const char *file)
{
    FILE *fp;
    char *ret = (char*)0, *nl;
    size_t len = 0;

    if ((fp = fopen(file, "r")) == (FILE*)0)
	goto err;
    if (getline(&ret, &len, fp) < 0)
	goto out;
    if (len)
	ret[len-1] = '\0';
    if ((nl = strchr(ret, '\n')))
	*nl = '\0';
out:
    fclose(fp);
err:
    return ret;
}

/* Read and determine active attribute for tty */
static char *actattr(const char *tty)
{
    char *ret = (char*)0;
    char *path;

    if (!tty || *tty == '\0')
	goto err;

    if (asprintf(&path, "/sys/class/tty/%s/active", tty) < 0)
	goto err;

    if ((ret = oneline(path)) == (char*)0)
	goto out;
out:
    free(path);
err:
    return ret;
}

/* Read and determine device attribute for tty */
static dev_t devattr(const char *tty)
{
    unsigned int maj, min;
    dev_t dev = 0;
    char *path, *value;

    if (!tty || *tty == '\0')
	goto err;

    if (asprintf(&path, "/sys/class/tty/%s/dev", tty) < 0)
	goto err;

    if ((value = oneline(path)) == (char*)0)
	goto out;

    if (sscanf(value, "%u:%u", &maj, &min) == 2)
	dev = makedev(maj, min);
    free(value);
out:
    free(path);
err:
    return dev;
}

#undef USE_SYSFS    /* Can not determine preferred console */

/* main routine to fetch tty */
char * fetchtty(const pid_t pid, const pid_t ppid, dev_t * tty)
{
#ifdef USE_SYSFS
    char * attrib;
#endif
    char * name = NULL;
    dev_t dev = 0;
    int found = 0;
    DIR * ddev;
    FILE * fc;

    if ((fc = fopen("/proc/consoles", "r"))) {
	char fbuf[16];
	unsigned int maj, min;
	while ((fscanf(fc, "%*s %*s (%[^)]) %u:%u", &fbuf[0], &maj, &min) == 3)) {
	    if (!strchr(fbuf, 'E'))
		continue;
	    if (strchr(fbuf, 'C')) {
		dev = makedev(maj, min);
		break;
	    }
	}
	fclose(fc);
#ifdef USE_SYSFS
    } else if ((attrib = actattr("console"))) {
	char *words = attrib, *token;
	while ((token = strsep(&words, " \t\n"))) {
	    if (*token)
		 name = token;	/* At least this takes the last (preferred console??) */
	}
	dev = devattr(name);
	free(attrib);
	if (dev == makedev(TTY_MAJOR, 0)) {
	    if ((attrib = actattr(name))) {
		dev = devattr(attrib);
		free(attrib);
	    }
	}
#endif
    } else {
#ifdef TIOCGDEV
	unsigned int devnum = 0;
	if (ioctl (0, TIOCGDEV, &devnum) < 0) {
	    if (errno == EINVAL && !getenv("NOTIOCGDEV"))
		warn("Warning: the ioctl TIOCGDEV is not known by the kernel\n");
	}
	dev = (dev_t)devnum;
#else
	struct stat st;
	if (fstat(0, &st) == 0 && (major(st.st_rdev) == TTYAUX_MAJOR || st.st_rdev == makedev(TTY_MAJOR, 0))) {
	    unsigned int arg;
	    if (ioctl(0, TIOCMGET, &arg) == 0) {		/* serial line */
# if 0
		dev = makedev(TTY_MAJOR, 64);			/* 8250/16450/16550(A) ports */
# else
		dev = 0;
# endif
	    } else if (ioctl(0, KDGKBTYPE, &arg) == 0) {	/* virtual console */
		struct vt_stat vt;
		if (ioctl(0, VT_GETSTATE, &vt) == 0)
		    dev = makedev(TTY_MAJOR, (int)vt.v_active);
	    }
	}
#endif
    }
    if (!dev) {
	if (!(name = ttyname(0)) || !strcmp(name, "/dev/console"))
	    dev = fallback(pid, ppid);
	else {
	    name = realpath(name, NULL);
	    if (!name)
		error("fetchtty(): %m\n");
	    goto out;
	}
    }
    if (tty) *tty = dev;

    if (!(ddev = opendir("/dev")))
	error("can not opendir(/dev): %m\n");
    pushd("/dev");
    found = checkdev(&name, dev, ddev);
    popd();
    closedir(ddev);

    if (!name)
	goto out;

    if (!found) {
	free(name);
	name = (char*)0;
    }
out:
    return name;
}

/* Allocate a console */
static void consalloc(struct console *restrict cons, char * name)
{
    struct console *restrict tail = cons;
    struct console *restrict newc;
    int flags;

    if (posix_memalign((void*)&newc, sizeof(void*), alignof(typeof(struct console))) != 0)
	error("memory allocation: %m\n");

    newc->next = (struct console*)0;
    newc->tty = name;
    if ((newc->fd = open(newc->tty, O_WRONLY|O_NONBLOCK|O_NOCTTY)) < 0) {
	warn("can not open %s: %m\n", newc->tty);
	free(newc);
	return;
    }
    tail->next = newc;
    tail->tlock = 0;
    tail->max_canon = _POSIX_MAX_CANON;
    memset(&tail->ltio, 0, sizeof(tail->ltio));
    memset(&tail->otio, 0, sizeof(tail->otio));
    tail = tail->next;

    if ((flags = fcntl(newc->fd, F_GETFL)) < 0)
	warn("can not get terminal flags of %s: %m\n", newc->tty);

    flags &= ~(O_NONBLOCK);
    flags |=   O_NOCTTY;
    if (fcntl(newc->fd, F_SETFL, flags) < 0)
	warn("can not set terminal flags of %s: %m\n", newc->tty);
}

/* Do we have some more system console around? */
void secondtty(struct console *restrict cons, const dev_t tty)
{
    char buf[1024], *ptr, *shcmp;
    char *attrib;
    char *name = (char*)0;
    int fd = -1, len;
    struct stat st;
    dev_t rdev;
    DIR * ddev;
    FILE * fc;

    if (!cons || !cons->tty)
	error("no primary console\n");

    if (stat(cons->tty, &st) < 0) {
	warn("can not stat /dev/%s: %m\n", cons->tty);
	st.st_rdev = 0;
    }
    rdev = st.st_rdev;

    if ((fc = fopen("/proc/consoles", "r"))) {
	unsigned int maj, min;
	char fbuf[16];

	if (!(ddev = opendir("/dev")))
	    error("can not opendir(/dev): %m\n");
	pushd("/dev");

	while ((fscanf(fc, "%*s %*s (%[^)]) %u:%u", &fbuf[0], &maj, &min) == 3)) {
	    dev_t dev;

	    if (!strchr(fbuf, 'E'))
		continue;
	    if (strchr(fbuf, 'C'))
		continue;
	    dev = makedev(maj, min);

	    if (dev == tty)
		continue;
	    if (dev == rdev)
		continue;

	    if (dev == makedev(TTY_MAJOR, 0))	/* Avoid endless loop */
		continue;

	    if (!checkdev(&name, dev, ddev) || !name) {
		if (name) {
		    free(name);
		    name = (char*)0;
		}
		continue;
	    }

	    consalloc(cons, name);
	}

	popd();
	closedir(ddev);

	fclose(fc);
	return;
    }

    if ((attrib = actattr("console"))) {
	char *words = attrib, *token;

	if (!(ddev = opendir("/dev")))
	    error("can not opendir(/dev): %m\n");
	pushd("/dev");

	while ((token = strsep(&words, " \t\n"))) {
	    dev_t dev;

	    if (*token == '\0')
		continue;

	    dev = devattr(token);

	    if (dev == makedev(TTY_MAJOR, 0)) {
		char *tmp = actattr(token);
		if (!tmp)
		    continue;
		dev = devattr(tmp);
		free(tmp);
	    }

	    if (dev == tty)
		continue;
	    if (dev == rdev)
		continue;

	    if (dev == makedev(TTY_MAJOR, 0))	/* Avoid endless loop */
		continue;

	    if (!checkdev(&name, dev, ddev) || !name) {
		if (name) {
		    free(name);
		    name = (char*)0;
		}
		continue;
	    }

	    consalloc(cons, name);
	}
	popd();
	closedir(ddev);
	free(attrib);

	return;
    }

    if ((fd = open("/proc/cmdline", O_RDONLY|O_NOCTTY)) < 0) {
	warn("can not open /proc/cmdline\n");
	return;
    }

    if ((len = read(fd, buf, sizeof(buf) - 1)) < 0) {
	warn("can not read /proc/cmdline\n");
	close(fd);
	return;
    }
    close(fd);

    if (len == 0)
	return;

    shcmp = cons->tty;
    if (!strncmp(shcmp, "/dev/", 5))
	shcmp += 5;

    /*
     * Check for e.g. /dev/tty[1-9] which is equal to /dev/tty0
     */
    if (!strncmp(shcmp, "tty", 3)) {
	size_t len = strspn(shcmp + 3, "123456789");
	if (strlen(shcmp) == len + 3)
	    shcmp   = "tty0";
    }

    if (!(ddev = opendir("/dev")))
	error("can not opendir(/dev): %m\n");
    pushd("/dev");

    ptr = &buf[len];
    while (ptr >= &buf[0]) {
	char *res = (char*)0;
#ifdef TIOCGDEV
	unsigned int devnum;
#else
	struct vt_stat vt;
#endif
	dev_t dev;
	int fd;

	if (*ptr == ',' || *ptr == ' ' || *ptr == '\t' || *ptr == '\r' || *ptr == '\n') {
	    *ptr-- = 0;
	    continue;
	}
	if (*ptr == 'c' && !strncmp(ptr, "console=", 8)) {
	    char * console  = ptr + 8;

	    if (!strncmp(console, "/dev/", 5))
		console += 5;

	    if (!strcmp(console, "brl"))
		console += 4;

	    /*
	     * Compare known console tty with that of the kernel command
	     * line.  If already known skip this console tty and search
	     * for the next one.
	     */
	    if (strcmp(shcmp, console))
		res = console;		/* New device not identical to tty */
	}
	ptr--;

	if (!res)
	    continue;

	/* Open second console e.g. /dev/tty0 */
	if ((fd = open(res, O_RDWR|O_NONBLOCK|O_NOCTTY)) < 0)
	    continue;

	/*
	 * We do this because if we would write out the buffered
	 * messages to e.g. /dev/tty0 after this we would (re)read
	 * those and buffer them again which leads to an endless loop.
	 */
#ifdef TIOCGDEV
	if (ioctl (fd, TIOCGDEV, &devnum) < 0) {
	    if (errno == EINVAL && !getenv("NOTIOCGDEV"))
		warn("Warning: the ioctl TIOCGDEV is not known by the kernel\n");
	    close(fd);
	    continue;
	}
	dev = (dev_t)devnum;
#else
	if (fstat(fd, &st) < 0) {
	    close(fd);
	    continue;
	}
	dev = st.st_rdev;
	if (dev == makedev(TTY_MAJOR, 0)) {
	    if (ioctl(fd, VT_GETSTATE, &vt) < 0) {
		if (errno == EINVAL)
		    warn("Warning: the ioctl VT_GETSTATE is not known by the kernel\n");
		close(fd);
		continue;
	    }
	    dev = makedev(TTY_MAJOR, (int)vt.v_active);
	}
#endif
	close(fd);

	if (dev == tty)
	    continue;
	if (dev == rdev)
	    continue;

	if (dev == makedev(TTY_MAJOR, 0))	/* Avoid endless loop */
	    continue;

	if (!checkdev(&name, dev, ddev) || !name) {
	    if (name) {
		free(name);
		name = (char*)0;
	    }
	    continue;
	}

	consalloc(cons, name);
    }

    popd();
    closedir(ddev);
}
