/*
 * showconsole.c
 *
 * Copyright 2000 Werner Fink, 2000 SuSE GmbH Nuernberg, Germany.
 *
 * This source is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <sys/types.h> /* Defines the macros major and minor */
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include "libconsole.h"

/*
 * Internal logger
 */
static char *myname = NULL;
static void _logger (const char *fmt, va_list ap)
{
    char buf[strlen(myname)+2+strlen(fmt)+1];
    strcat(strcat(strcpy(buf, myname), ": "), fmt);
    vfprintf(stderr, buf, ap);
    return;
}

/*
 * Cry and exit.
 */
void error (const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    _logger(fmt, ap);
    va_end(ap);
    exit (1);
}

/*
 * Warn the user.
 */
void warn (const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    _logger(fmt, ap);
    va_end(ap);
    return;
}

/*
 * Now do the job
 */
int main(int argc, char *argv[])
{
    const char* opt = argv[1];
    char * tty = NULL, numeric = 0;
    myname = basename(*argv);

    if (!strcmp(myname, "setconsole")) {
	char *curtty = ttyname(0);
	int fd, isconsole = 0;

	if (curtty && (isconsole = (strcmp("/dev/console", curtty) == 0)))
	    fd = 0;
	else
	    fd = open("/dev/console", O_RDWR|O_NOCTTY);

	if (fd < 0)
	    error("can not open console: %m\n");

	(void)ioctl(fd, TIOCCONS, NULL);        /* Undo any current map if any */
	close(fd);

	if (argc < 2)
	    goto out;

	if (argc > 2)
	    error("Usage: %s [-r | <tty> ]\n", myname);

	if (opt && *opt++ == '-' && *opt++ == 'r' && *opt == '\0')
	    goto out;

	if ((fd = open(argv[1], O_WRONLY|O_NOCTTY)) < 0)
	    error("can not open %s: %m\n", argv[1]);

	if (!isatty(fd))
	    error("%s is not a tty.\n", argv[1]);

	if (ioctl(fd, TIOCCONS, NULL) < 0)
	    error("can not set console device: %m\n");

	goto out;
    }

    if (argc == 2) {
	if (opt && *opt++ == '-' && *opt++ == 'n' && *opt == '\0')
	    numeric++;
	else
	    error("Usage: %s [-n]\n", myname);
    } else if (argc > 2)
	error("Usage: %s [-n]\n", myname);

    if (numeric) {
	dev_t dev = 0;
	(void)fetchtty(getpid(), getppid(), &dev);

	if (dev)
	    printf("%u %u\n", major(dev), minor(dev));
	else {
	    error("real tty unknown\n");
	    fprintf(stderr, "real tty unknown\n");
	}
	goto out;
    }

    if ((tty = fetchtty(getpid(), getppid(), NULL)))
	printf("%s\n", tty);
    else {
	error("real tty unknown\n");
	fprintf(stderr, "real tty unknown\n");
    }
out:
    return 0;
}
